import copy
import typing as tp

from satella.coding.decorators.decorators import wraps
from satella.coding.recast_exceptions import rethrow_as
from .iterators import n_th
from ..typing import T, Iteratable, NoArgCallable, Predicate


def infinite_iterator(returns: tp.Optional[T] = None,
                      return_factory: tp.Optional[NoArgCallable[T]] = None) -> tp.Iterator[T]:
    """
    Return an infinite number of objects.

    :param returns: object to return. Note that this will be this very object, it will
        not be copied.
    :param return_factory: a callable that takes 0 args and returns an element to return.
    :return: an infinite iterator of provided values
    """
    while True:
        if returns is None:
            if return_factory is None:
                yield None
            else:
                yield return_factory()
        else:
            yield returns


def make_list(element: T, n: int, deep_copy: bool = False) -> tp.List[T]:
    """
    Make a list consisting of n times element. Element will be copied via
    copy.copy before adding to list.

    :param element: element
    :param n: times to repeat the element
    :param deep_copy: whether to use copy.deepcopy instead of copy.copy
    :return: list of length n
    """
    output = []

    if deep_copy:
        copy_op = copy.deepcopy
    else:
        copy_op = copy.copy

    for _ in range(n):
        output.append(copy_op(element))
    return output


# shamelessly copied from
# https://medium.com/better-programming/is-this-the-last-element-of-my-python-for-loop-784f5ff90bb5
def is_last(lst: Iteratable) -> tp.Iterator[tp.Tuple[bool, T]]:
    """
    Return every element of the list, alongside a flag telling is this the last element.

    Use like:

    >>> for is_last, element in is_last(my_list):
    >>>     if is_last:
    >>>         ...

    :param lst: list to iterate thru
    :return: a p_gen returning (bool, T)

    Note that this returns a nice, O(1) iterator.
    """
    iterable = iter(lst)
    ret_var = next(iterable)
    for val in iterable:
        yield False, ret_var
        ret_var = val
    yield True, ret_var


def add_next(lst: Iteratable,
             wrap_over: bool = False,
             skip_last: bool = False) -> tp.Iterator[tp.Tuple[T, tp.Optional[T]]]:
    """
    Yields a 2-tuple of given iterable, presenting the next element as second element of the tuple.

    The last element will be the last element alongside with a None, if wrap_over is False, or the
    first element if wrap_over was True

    Example:

    >>> list(add_next([1, 2, 3, 4, 5])) == [(1, 2), (2, 3), (3, 4), (4, 5), (5, None)]
    >>> list(add_next([1, 2, 3, 4, 5], True)) == [(1, 2), (2, 3), (3, 4), (4, 5), (5, 1)]

    :param lst: iterable to iterate over
    :param wrap_over: whether to attach the first element to the pair of the last element instead
        of None
    :param skip_last: if this is True, then last element, alongside with a None, won't be output
    """
    iterator = iter(lst)
    try:
        first_val = prev_val = next(iterator)
    except StopIteration:
        return
    for val in iterator:
        yield prev_val, val
        prev_val = val
    if wrap_over:
        yield prev_val, first_val
    else:
        if not skip_last:
            yield prev_val, None


def half_cartesian(seq: tp.Iterable[T],
                   include_same_pairs: bool = True) -> tp.Iterator[tp.Tuple[T, T]]:
    """
    Generate half of the Cartesian product of both sequences.

    Useful when you have a commutative operation that you'd like to execute on both elements
    (eg. checking for collisions).

    Example:

    >>> list(half_cartesian([1, 2, 3], [1, 2, 3])) == \
    >>>     [(1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (3, 3)]

    :param seq: The sequence
    :param include_same_pairs: if True, then pairs returning two of the same objects will be
        returned. For example, if False, the following will be true:

    >>> list(half_cartesian([1, 2, 3], [1, 2, 3], include_same_pairs=False)) == \
    >>>     [(1, 2), (1, 3), (2, 3)]

    """
    for i, elem1 in enumerate(seq):
        for j, elem2 in enumerate(seq):
            if include_same_pairs:
                if j >= i:
                    yield elem1, elem2
            else:
                if j > i:
                    yield elem1, elem2


def group_quantity(length: int, seq: Iteratable) -> tp.Iterator[tp.List[T]]:
    """
    Slice an iterable into lists containing at most len entries.

    Eg.

    >>> assert list(group_quantity(3, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])) == [[1, 2, 3], [4, 5, 6],
    >>>                                                                     [7, 8, 9], [10]]

    This correctly detects sequences, and uses an optimized variant via slicing if
    a sequence is passed.

    You can safely pass ranges

    :param length: p_len for the returning sequences
    :param seq: sequence to split
    """
    if isinstance(seq, tp.Sequence) and not isinstance(seq, range):
        i = 0
        while i < len(seq):
            yield seq[i:i + length]
            i += length
    else:
        entries = []
        for elem in seq:
            if len(entries) == length:
                yield entries
                entries = [elem]
            else:
                entries.append(elem)

        if entries:
            yield entries


def filter_out_nones(y: tp.Sequence[T]) -> tp.List[T]:
    """
    Return all elements, as a list, that are not None

    :param y: a sequence of items
    :return: a list of all subelements, in order, that are not None
    """
    output = []
    for item in y:
        if item is not None:
            output.append(item)
    return output


def filter_out_false(y: tp.Sequence[T]) -> tp.List[T]:
    """
    Return all elements, as a list, that are True

    :param y: a sequence of items
    :return: a list of all subelements, in order, that are not None
    """
    output = []
    for item in y:
        if item:
            output.append(item)
    return output


@rethrow_as(IndexError, ValueError)
def index_of_max(seq: tp.Sequence[T]) -> int:
    """
    Return the index of the maximum element

    :param seq: sequence to examine
    :return: index of the maximum element
    :raise ValueError: sequence was empty
    """
    max_index = 0
    max_elem = seq[0]
    for i, elem in enumerate(seq):
        if elem > max_elem:
            max_index = i
            max_elem = elem
    return max_index


def index_of(predicate: Predicate, seq: tp.Sequence[T]) -> int:
    """
    Return an index of first met element that calling predicate on it returns True

    :param predicate: predicate to apply
    :param seq: sequence to examine
    :return: index of the element
    :raises ValueError: if no element found
    """
    i = 0
    for elem in seq:
        if predicate(elem):
            return i
        i += 1
    raise ValueError('Element not found')


class Multirun:
    """
    A class to launch the same operation on the entire sequence.

    Consider:

    >>> class Counter:
    >>>     def __init__(self, value=0):
    >>>         self.count = value
    >>>     def add(self, v):
    >>>         self.count += 1
    >>>     def __eq__(self, other):
    >>>          return self.count == other.count
    >>>     def __iadd__(self, other):
    >>>         self.add(other)
    >>> a = [Counter(), Counter()]

    The following:

    >>> for b in a:
    >>>     b.add(2)

    Can be replaced with

    >>> Multirun(a).add(2)

    And the following:

    >>> for b in a:
    >>>     b += 3

    With this

    >>> b = Mulirun(a)
    >>> b += 3

    Furthermore note that:

    >>> Multirun(a).add(2) == [Counter(2), Counter(2)]

    :param sequence: sequence to execute these operations for
    :param dont_return_list: the operation won't return a list if this is True
    """
    __slots__ = ('sequence', 'dont_return_list')

    def __bool__(self) -> bool:
        return bool(self.sequence)

    def __init__(self, sequence: tp.Iterable, dont_return_list: bool = False):
        self.sequence = sequence
        self.dont_return_list = dont_return_list

    def __iter__(self):
        return iter(self.sequence)

    def __getattr__(self, item):
        def inner(*args, **kwargs):
            if not self.dont_return_list:
                results = []
                for element in self:
                    getattr(element, item)(*args, **kwargs)
                    results.append(element)
                return results
            else:
                for element in self:
                    getattr(element, item)(*args, **kwargs)

        # Take care: the array might just be empty...
        try:
            fun = getattr(n_th(self), item)
            inner = wraps(fun)(inner)
        except IndexError:
            pass

        return inner

    def __iadd__(self, other):
        for element in self:
            element += other
        return self

    def __isub__(self, other):
        for element in self:
            element -= other
        return self

    def __imul__(self, other):
        for element in self:
            element *= other
        return self

    def __itruediv__(self, other):
        for element in self:
            element /= other
        return self

    def __ifloordiv__(self, other):
        for element in self:
            element //= other
        return self

    def __ilshift__(self, other):
        for element in self:
            element <<= other
        return self

    def __irshift__(self, other):
        for element in self:
            element >>= other
        return self

    def __ipow__(self, other):
        for element in self:
            element **= other
        return self



import os
import sys
import warnings


def is_running_as_root() -> bool:
    """
    Is this process running as root?

    Checks whether effective UID is 0

    :return: bool
    :raises OSError: called on Windows!
    """

    if sys.platform.startswith('win'):
        raise OSError('Routine unavailable on Windows!')

    return os.geteuid() == 0


def suicide(kill_entire_pg: bool = True) -> None:
    """
    Kill self.

    :param kill_entire_pg: whether to kill entire PG if a session leader. Won't work on Windows.
    """
    my_pid = os.getpid()
    if sys.platform.startswith('win'):
        if kill_entire_pg:
            warnings.warn('Windows does not support process leader groups', RuntimeWarning)
        os.kill(my_pid, -9)

    if kill_entire_pg and os.getpgid(0) == my_pid:
        os.killpg(my_pid, 9)
    else:
        os.kill(my_pid, 9)



import typing as tp
from abc import ABCMeta, abstractmethod

NoneType = None.__class__
T = tp.TypeVar('T')
Iteratable = tp.Union[tp.Iterator[T], tp.Iterable[T]]
U = tp.TypeVar('U')
V = tp.TypeVar('V')
K = tp.TypeVar('K', covariant=tp.Hashable)
KVTuple = tp.Tuple[K, V]
Number = tp.Union[int, float]
NoArgCallable = tp.Callable[[], T]
Predicate = tp.Callable[[T], bool]

ExceptionClassType = tp.Type[Exception]


class ClassComparable(metaclass=ABCMeta):
    @abstractmethod
    def __gt__(self, other: tp.Any) -> bool:
        ...

    @abstractmethod
    def __lt__(self, other: tp.Any) -> bool:
        ...

    @abstractmethod
    def __eq__(self, other: tp.Any) -> bool:
        ...


Comparable = tp.TypeVar('Comparable', bound=ClassComparable)

try:
    from typing import Protocol
except ImportError:
    Protocol = tp.Generic


class Appendable(Protocol[T]):
    def append(self, item: T) -> None:
        ...


__all__ = ['Iteratable', 'T', 'U', 'V', 'K', 'Number', 'ExceptionClassType',
           'NoArgCallable', 'Appendable', 'Predicate', 'KVTuple',
           'Comparable']



import copy
import time
import typing as tp

from .base import LeafMetric, MetricLevel
from .registry import register_metric
from ..data import MetricDataCollection, MetricData


@register_metric
class CallableMetric(LeafMetric):
    """
    A metric whose value at any given point in time is the result of it's callable.

    :param value_getter: a callable() that returns a float - the current value of this metric.
        It should be easy and cheap to compute, as this callable will be called each time
        a snapshot of metric state is requested
    """
    CLASS_NAME = 'callable'

    __slots__ = ('callable', 'labeled_metrics')

    def __init__(self, name, root_metric: 'Metric' = None,
                 metric_level: tp.Optional[MetricLevel] = None,
                 labels: tp.Optional[dict] = None, internal: bool = False,
                 value_getter: tp.Optional[tp.Callable[[], float]] = None, *args, **kwargs):
        super().__init__(name, root_metric, metric_level, labels, internal, *args, **kwargs)
        self.callable = value_getter
        self.labeled_metrics = []

    def register_labeled_metric(self, labeled_metric):
        self.labeled_metrics.append(labeled_metric)

    def _handle(self, *args, **kwargs) -> None:
        raise TypeError('You are not supposed to call this!')

    def to_metric_data(self) -> MetricDataCollection:
        mdc = MetricDataCollection()
        for labeled_metric in self.labeled_metrics:
            labels = copy.copy(self.labels)
            labels.update(labeled_metric.labels)
            mdc += MetricData(self.name, labeled_metric.callable(), labels, time.time(),
                              self.internal)

        if self.callable:
            mdc += MetricData(self.name, self.callable(), self.labels, time.time(), self.internal)

        return mdc



import typing as tp

from ..decorators.decorators import wraps


# noinspection PyPep8Naming
def Singleton(cls):
    """
    Make a singleton out of decorated class.

    Usage:

    >>> @Singleton
    >>> class MyClass:
    >>>     ...
    """

    cls.__new_old__ = cls.__new__

    @wraps(cls.__new__)
    def singleton_new(cls, *args, **kw):
        it = cls.__dict__.get('__it__')
        if it is not None:
            return it

        cls.__it__ = it = cls.__new_old__(cls)
        it.__init_old__(*args, **kw)
        return it

    cls.__new__ = singleton_new
    cls.__init_old__ = cls.__init__
    cls.__init__ = wraps(cls.__init__)(
        lambda self, *args, **kwargs: object.__init__(self))

    return cls


# noinspection PyPep8Naming
def SingletonWithRegardsTo(num_args: int):
    """
    Make a memoized singletion depending on the arguments.

    A dictionary is made (first N arguments => class instance) and such is returned.
    Please take care to ensure that a tuple made out of first num_args can be used as a dictionary
    key (ie. is both hashable and __eq__-able).

    Usage:

    >>> @SingletonWithRegardsTo(num_args=1)
    >>> class MyClass:
    >>>     def __init__(self, device_id: str):
    >>>         ...
    >>> a = MyClass('dev1')
    >>> b = MyClass('dev2')
    >>> c = MyClass('dev1')
    >>> assert a is c
    >>> assert b is not c
    """

    def inner(cls):

        cls.__new_old__ = cls.__new__

        @wraps(cls.__new__)
        def singleton_new(cls, *args, **kw):
            it = cls.__dict__.get('__it__')
            if it is None:
                it = cls.__it__ = {}

            key = args[:num_args]
            if key in it:
                return it[key]

            inst = it[key] = cls.__new_old__(cls)
            inst.__init_old__(*args, **kw)
            return inst

        cls.__new__ = singleton_new
        cls.__init_old__ = cls.__init__
        cls.__init__ = wraps(cls.__init__)(
            lambda self, *args, **kwargs: object.__init__(self))

        return cls

    return inner


def get_instances_for_singleton(x) -> tp.List[tp.Tuple]:
    """
    Obtain a list of arguments for which singletons exists for given
    class decorated with SingletonWithRegardsTo

    :param x: a class decorated with SingletonWithRegardsTo
    :return: a list of arguments
    """
    return list(x.__it__)


def delete_singleton_for(x, *args) -> None:
    """
    Delete singleton for given arguments in a class decorated with SingletonWithRegardsTo

    :param x: class decorated with SingletonWithRegardsTo
    :param args: arguments used in the constructor
    """
    del x.__it__[args]



import typing as tp
from collections import namedtuple


def _nth(it, n):
    return [x[n] for x in it]


def _adjust(q):
    var, type_ = q
    if type_ in (None, 'self') or isinstance(var, type_):
        return var
    return type_(var)


def typednamedtuple(cls_name: str, *arg_name_type: type) -> tp.Type[tp.Tuple]:
    """
    Returns a new subclass of tuple with named fields.
    Fields will be coerced to type passed in the pair, if they are not already
    of given type.

    Parameters are tuples of (field name, class/constructor as callable/1)

    For example:

    >>> tnt = typednamedtuple('tnt', ('x', float), ('y', float))
    >>> a = tnt('5.0', y=2)

    a.x is float, a.y is float too
    """

    fieldnames = []
    typeops = []
    mapping = {}

    for name, type_ in arg_name_type:
        fieldnames.append(name)
        typeops.append(type_)
        mapping[name] = type_

    MyCls = namedtuple(cls_name, fieldnames)

    class Wrapper(MyCls):
        __doc__ = MyCls.__doc__
        __name__ = MyCls.__name__

        def __new__(cls, *args, **kwargs):

            nargs = list(map(_adjust, zip(args, typeops[:len(args)])))

            for next_field_name in fieldnames[len(nargs):]:
                try:
                    nargs.append(_adjust((kwargs.pop(next_field_name), mapping[next_field_name])))
                except KeyError:
                    raise TypeError('Field %s not given', next_field_name)

            if len(kwargs) > 0:
                raise TypeError('Too many parameters')

            return MyCls.__new__(MyCls, *nargs)

    return Wrapper



from .average import RollingArithmeticAverage
from .choose import choose, choose_one
from .iterators import infinite_counter, take_n, is_instance, skip_first, zip_shifted, \
    stop_after, iter_dict_of_list, shift, other_sequence_no_longer_than, count, even, \
    odd, n_th, smart_enumerate, smart_zip, unique, ConstruableIterator, walk, length, map_list, \
    is_empty, IteratorListAdapter, enumerate2, to_iterator, ListWrapperIterator, try_close, \
    f_range
from .sequences import is_last, add_next, half_cartesian, group_quantity, Multirun, make_list, \
    infinite_iterator, filter_out_false, filter_out_nones, index_of, index_of_max

__all__ = ['choose', 'choose_one', 'infinite_counter', 'take_n', 'is_instance', 'is_last',
           'add_next', 'ConstruableIterator', 'walk', 'length', 'smart_zip',
           'half_cartesian', 'skip_first', 'zip_shifted', 'stop_after', 'group_quantity',
           'iter_dict_of_list', 'shift', 'other_sequence_no_longer_than', 'count', 'n_th',
           'even', 'odd', 'Multirun', 'smart_enumerate', 'unique', 'map_list', 'is_empty',
           'make_list', 'RollingArithmeticAverage', 'index_of', 'index_of_max',
           'IteratorListAdapter', 'enumerate2', 'infinite_iterator', 'to_iterator',
           'filter_out_false', 'filter_out_nones', 'ListWrapperIterator', 'try_close',
           'f_range']



import collections
import copy
import weakref
import time
import typing as tp

from satella.coding.deleters import DictDeleter
from satella.coding.typing import T, NoArgCallable


class CancellableCallback:
    """
    A callback that you can cancel.

    Useful for event-driven software that looks through lists of callbacks and determines whether
    to delete them or further invalidate in some other way.

    If called, the function itself won't be called as well if this was cancelled. In this case
    a None will be returned instead of the result of callback_fun()

    This short circuits __bool__ to return not .cancelled.

    Hashable and __eq__-able by identity.

    :param callback_fun: function to call

    :ivar cancelled: whether this callback was cancelled (bool)
    """
    __slots__ = ('cancelled', 'callback_fun')

    def __bool__(self) -> bool:
        return not self.cancelled

    def __hash__(self) -> int:
        return hash(id(self))

    def __init__(self, callback_fun: tp.Callable):
        self.callback_fun = callback_fun
        self.cancelled = False

    def __call__(self, *args, **kwargs):
        if not self.cancelled:
            return self.callback_fun(*args, **kwargs)

    def cancel(self) -> None:
        """
        Cancel this callback.
        """
        self.cancelled = True


class CallableGroup(tp.Generic[T]):
    """
    This behaves like a function, but allows to add other functions to call
    when invoked, eg.

        c1 = Callable()

        c1.add(foo)
        c1.add(bar)

        c1(2, 3)

    Now both foo and bar will be called with arguments (2, 3). Their exceptions
    will be propagated.

    """
    __slots__ = ('callables', 'gather', 'swallow_exceptions',
                 '_has_cancellable_callbacks')

    def __init__(self, gather: bool = True, swallow_exceptions: bool = False):
        """
        :param gather: if True, results from all callables will be gathered
                       into a list and returned from __call__
        :param swallow_exceptions: if True, exceptions from callables will be
                                   silently ignored. If gather is set,
                                   result will be the exception instance
        """
        self.callables = collections.OrderedDict()  # type: tp.Dict[tp.Callable, bool]
        self.gather = gather  # type: bool
        self.swallow_exceptions = swallow_exceptions  # type: bool
        self._has_cancellable_callbacks = False

    @property
    def has_cancelled_callbacks(self) -> bool:
        """
        Check whether this has any
        :class:`~satella.coding.concurrent.CancellableCallback` instances and whether any
        of them was cancelled
        """
        if not self._has_cancellable_callbacks:
            return False
        for clb in self.callables:
            if isinstance(clb, CancellableCallback) and not clb:
                return True
        return False

    def remove_cancelled(self) -> None:
        """
        Remove it's entries that are CancelledCallbacks and that were cancelled
        """
        if not self.has_cancelled_callbacks:
            return

        with DictDeleter(self.callables) as dd:
            for callable_ in dd:
                if isinstance(callable_, CancellableCallback) and not callable_:
                    dd.delete()

    def add(self, callable_: tp.Union[CancellableCallback, NoArgCallable[T]],
            one_shot: bool = False):
        """
        Add a callable.

        Can be a :class:`~satella.coding.concurrent.CancellableCallback`, in that case
        method :meth:`~satella.coding.concurrent.CallableGroup.remove_cancelled` might
        be useful.

        :param callable_: callable
        :param one_shot: if True, callable will be unregistered after single call
        """
        if isinstance(callable_, CancellableCallback):
            self._has_cancellable_callbacks = True
        from ..structures.hashable_objects import HashableWrapper
        callable_ = HashableWrapper(callable_)
        if callable_ in self.callables:
            return
        self.callables[callable_] = one_shot

    def __call__(self, *args, **kwargs) -> tp.Optional[tp.List[T]]:
        """
        Run the callable. All registered callables will be called with
        passed arguments, so they should have the same arity.

        If callables raise, it will be passed through.

        As a side-effect, removes cancelled
        :class:`~satella.coding.concurrent.CancellableCallback` instances registered.

        :return: list of results if gather was set, else None
        """
        if self.has_cancelled_callbacks:
            self.remove_cancelled()

        callables = copy.copy(self.callables)
        self.callables.clear()

        results = []

        for call, one_shot in callables.items():
            try:
                q = call(*args, **kwargs)
            except Exception as e:
                if not self.swallow_exceptions:
                    raise  # re-raise
                q = e

            if self.gather:
                results.append(q)

            if not one_shot:
                self.add(call, one_shot)

        if self.gather:
            return results


class CallNoOftenThan:
    """
    A class that will ensure that calls to given callable are made no more often
    than some interval.

    Even if it's call is called more often than specified value, the callable just
    won't be called and None will be returned.

    :param interval: interval in seconds
    :param callable_: callable to call
    """
    __slots__ = ('interval', 'callable', 'last_called')

    def __init__(self, interval: float, callable_: tp.Callable):
        self.interval = interval  # type: float
        self.callable = callable_  # type: tp.Callable
        self.last_called = 0  # type: float

    def __call__(self, *args, **kwargs):
        if time.monotonic() - self.last_called >= self.interval:
            self.callable(*args, **kwargs)
            self.last_called = time.monotonic()



import enum
import time
import typing as tp

from satella.coding.decorators import for_argument
from ..data import MetricData, MetricDataCollection


class MetricLevel(enum.IntEnum):
    DISABLED = 1
    RUNTIME = 2
    DEBUG = 3
    INHERIT = 4

    def __ge__(self, other: 'MetricLevel') -> bool:
        return self.value >= other.value


# todo deprecated, remove in 3.0 maybe?
DISABLED = MetricLevel.DISABLED
RUNTIME = MetricLevel.RUNTIME
DEBUG = MetricLevel.DEBUG
INHERIT = MetricLevel.INHERIT


class Metric:
    """
    Container for child metrics. A base metric class, as well as the default metric.

    Switch levels by setting metric.level to a proper value

    :param enable_timestamp: append timestamp of last update to the metric
    :param internal: if True, this metric won't be visible in exporters
    """
    __slots__ = ('name', 'root_metric', 'internal', '_level', 'enable_timestamp',
                 'last_updated', 'children')

    CLASS_NAME = 'base'

    def get_fully_qualified_name(self):
        data = []
        metric = self
        while metric.root_metric is not None:
            data.append(metric.name)
            metric = metric.root_metric
        return '.'.join(reversed(data))

    def reset(self) -> None:
        """
        Delete all child metrics that this metric contains.

        Also, if called on root metric, sets the runlevel to RUNTIME
        """
        from satella.instrumentation import metrics
        if self.name == '':
            with metrics.metrics_lock:
                metrics.metrics = {}
                metrics.level = MetricLevel.RUNTIME
        else:
            with metrics.metrics_lock:
                metrics.metrics = {k: v for k, v in metrics.metrics.items() if
                                   not k.startswith(self.get_fully_qualified_name() + '.')}
                del metrics.metrics[self.get_fully_qualified_name()]
        self.children = []

    def __init__(self, name, root_metric: 'Metric' = None,
                 metric_level: tp.Optional[tp.Union[MetricLevel, int]] = None,
                 internal: bool = False,
                 *args, **kwargs):
        """When reimplementing the method, remember to pass kwargs here!"""
        self.name = name
        self.root_metric = root_metric
        self.internal = internal
        if metric_level is None:
            if self.name == '':
                metric_level = MetricLevel.RUNTIME
            else:
                metric_level = MetricLevel.INHERIT
        self._level = MetricLevel(metric_level)  # type: MetricLevel
        self.enable_timestamp = kwargs.get('enable_timestamp', False)
        self.last_updated = time.time() if self.enable_timestamp else None \
            # type: tp.Optional[float]

        assert not (
                self.name == '' and self.level == MetricLevel.INHERIT), \
            'Unable to set INHERIT for root metric!'
        self.children = []  # type: tp.List[Metric]

    def get_timestamp(self) -> tp.Optional[float]:
        """Return this timestamp, or None if no timestamp support is enabled"""
        return self.last_updated if self.enable_timestamp else None

    def __str__(self) -> str:
        return self.name

    @property
    def level(self) -> MetricLevel:
        metric = self
        while metric._level == MetricLevel.INHERIT:
            metric = metric.root_metric
        return metric._level

    @level.setter
    @for_argument(None, MetricLevel)
    def level(self, value: MetricLevel) -> None:
        assert not (
                value == MetricLevel.INHERIT and self.name == ''), \
            'Cannot set INHERIT for the root metric!'
        self._level = value

    def append_child(self, metric: 'Metric'):
        self.children.append(metric)

    def can_process_this_level(self, target_level: MetricLevel) -> bool:
        return self.level >= target_level

    def to_metric_data(self) -> MetricDataCollection:
        output = MetricDataCollection()
        for child in self.children:
            if child.level <= self.level:
                output += child.to_metric_data()
        output.prefix_with(self.name)

        if self.enable_timestamp:
            output.set_timestamp(self.last_updated)

        return output

    def _handle(self, *args, **kwargs) -> None:
        """
        To be overridden!

        The right place to process your data, after it's level was verified by :meth:`Metric.handle`
        """
        raise TypeError('This is a container metric!')

    @for_argument(None, MetricLevel)
    def handle(self, level: tp.Union[int, MetricLevel], *args, **kwargs) -> None:
        if self.can_process_this_level(level):
            if self.enable_timestamp:
                self.last_updated = time.time()
            self._handle(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self.handle(MetricLevel.DEBUG, *args, **kwargs)

    def runtime(self, *args, **kwargs):
        self.handle(MetricLevel.RUNTIME, *args, **kwargs)


class LeafMetric(Metric):
    """
    A metric capable of generating only leaf entries.

    You cannot hook up any children to a leaf metric.
    """
    __slots__ = ('labels',)

    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None,
                 labels: tp.Optional[dict] = None, internal: bool = False, *args, **kwargs):
        super().__init__(name, root_metric, metric_level, internal, *args, **kwargs)
        self.labels = labels or {}
        assert '_timestamp' not in self.labels, 'Cannot make a label called _timestamp!'

    def to_metric_data(self) -> MetricDataCollection:
        return MetricDataCollection(
            MetricData(self.name, None, self.labels, internal=self.internal))

    def append_child(self, metric: 'Metric'):
        raise TypeError('This metric cannot contain children!')


class EmbeddedSubmetrics(LeafMetric):
    """
    A metric that can optionally accept some labels in it's handle, and this will be counted as a
    separate metric.
    For example:

    >>> metric = getMetric('root.test.IntValue', 'int', enable_timestamp=False)
    >>> metric.handle(2, label='key')
    >>> metric.handle(3, label='value')

    If you try to inherit from it, refer to :py:class:`.simple.IntegerMetric` to see how to do it.
    All please pass all the arguments received from child class into this constructor, as this
    constructor actually stores them!
    Refer to :py:class:`.cps.ClicksPerTimeUnitMetric` on how to do that.
    """
    __slots__ = ('args', 'kwargs', 'embedded_submetrics_enabled', 'children_mapping')

    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None,
                 labels: tp.Optional[dict] = None, internal: bool = False, *args, **kwargs):
        super().__init__(name, root_metric, metric_level, labels, internal, *args, **kwargs)
        self.args = args  # type: tp.List
        self.kwargs = kwargs  # type: tp.Dict
        self.embedded_submetrics_enabled = False  # type: bool
        self.children_mapping = {}  # type: tp.Dict[tp.Any, Metric]
        self.last_updated = time.time()  # type: float

    def _handle(self, *args, **labels):
        if self.enable_timestamp:
            self.last_updated = time.time()

        key = tuple(sorted(labels.items()))
        if key:
            self.embedded_submetrics_enabled = True
        else:
            return

        if key in self.children_mapping:
            # noinspection PyProtectedMember
            self.children_mapping[key]._handle(*args)
        else:
            clone = self.clone(labels)
            self.children_mapping[key] = clone
            self.children.append(clone)
            # noinspection PyProtectedMember
            self.children_mapping[key]._handle(*args)

    def to_metric_data(self) -> MetricDataCollection:
        if self.embedded_submetrics_enabled:
            v = MetricDataCollection()
            for child in self.children:
                if child.level <= self.level:
                    v = v + child.to_metric_data()
            return v
        else:
            return super().to_metric_data()

    def get_specific_metric_data(self, labels: dict) -> MetricDataCollection:
        """
        Return a MetricDataCollection for a child with given labels
        """
        key = tuple(sorted(labels.items()))
        return self.children_mapping[key].to_metric_data()

    def clone(self, labels: dict) -> 'LeafMetric':
        """
        Return a fresh instance of this metric, with it's parent being set to this metric
        and having a particular set of labels, and being of level INHERIT.
        """

        return self.__class__(self.name, self, MetricLevel.INHERIT, *self.args, labels=labels,
                              **self.kwargs)


from .registry import register_metric
register_metric(Metric)



import typing as tp

from satella.coding.typing import K


def hashables_to_int(words: tp.List[K]) -> tp.Dict[K, int]:
    """
    Assign each hashable an integer, starting from 0, and return the resulting mapping

    :param words: a list of hashables
    :return: a dictionary keyed by hashable and values are the assigned integers
    """
    dictionary = {}
    i = 0
    for word in set(words):
        dictionary[word] = i
        i += 1
    return dictionary



import enum


class ComparableEnum(enum.Enum):
    """
    An enum whose compare will try to convert value you compare it against
    to it's instance. Handly for writing code like:

    >>> a = 'test'
    >>> class Enum(ComparableEnum):
    >>>     A = 'test'
    >>> assert Enum.A == a

    Comparison order doesn't matter, so either are True:

    >>> Enum.A == 'test'
    >>> 'test' == Enum.A

    Note, however, that following won't work:

    >>> 'test' in (Enum.A, )

    You can even compare enums across classes

    >>> class A(ComparableEnum):
    >>>     A = 1
    >>> class B(ComparableEnum):
    >>>     A = 1
    >>> assert A.A == B.A
    """

    def __eq__(self, other) -> bool:
        if isinstance(other, enum.Enum) and not isinstance(other, self.__class__):
            other = other.value

        if not isinstance(other, self.__class__):
            try:
                return self.__class__(other) == self
            except ValueError:      # other is not a correct member of this class!
                return False
        else:
            return super().__eq__(other)


class HashableIntEnum(enum.IntEnum):
    """
    An enum.IntEnum that implements hashability, stemming from it's values, as well
    as hashability
    """

    def __hash__(self) -> int:
        return hash(self.value)


class ComparableIntEnum(HashableIntEnum):
    """
    An enum.IntEnum that implements comparision, stemming from it's values, as well
    as hashability
    """

    def __hash__(self) -> int:
        return hash(self.value)

    def __lt__(self, other: 'ComparableIntEnum') -> bool:
        return self.value < other.value

    def __le__(self, other: 'ComparableIntEnum') -> bool:
        return self.value <= other.value

    def __gt__(self, other: 'ComparableIntEnum') -> bool:
        return self.value > other.value

    def __ge__(self, other: 'ComparableIntEnum') -> bool:
        return self.value >= other.value

    def __eq__(self, other: 'ComparableIntEnum') -> bool:
        return self.value == other.value



import itertools
import typing as tp

from satella.coding.typing import T, Predicate
from satella.exceptions import PreconditionError
from .decorators import wraps
from .arguments import for_argument
from ..misc import source_to_function

Expression = tp.NewType('Expression', str)
Condition = tp.Union[Predicate[T], Expression]


# noinspection PyPep8Naming
def _TRUE(x):
    return True


def precondition(*t_ops: Condition, **kw_opts: Condition):
    """
    Check that a precondition happens for given parameter.

    You can do it like this:

    >>> @precondition(lambda x: x == 1)
    >>> def return_two(x):
    >>>     return x*2

    or

    >>> @precondition('x == 1')
    >>> def return_two(x):
    >>>     ..

    You can use all standard locals in precondition.

    You function call will return a PreconditionError (subclass of
    ValueError) if a precondition fails.

    A precondition of None will always be true.

    Keyword arguments are supported as well. Note that precondition for them will be checked
    only if they are passed, so make your default arguments obey the precondition, because
    it won't
    be checked if the default value is used.
    """

    tn_ops = []

    for t_op in t_ops:
        if t_op is None:
            precond_ = _TRUE
        else:
            precond_ = source_to_function(t_op)

        tn_ops.append(precond_)

    kw_ops = {}
    for kwarg_, value in kw_opts.items():
        if value is None:
            precond_ = _TRUE
        elif isinstance(value, str):
            precond_ = source_to_function(value)
        else:
            precond_ = value
        kw_ops[kwarg_] = precond_

    from satella.coding.recast_exceptions import rethrow_as

    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            assert len(args) >= len(tn_ops), 'More preconditions than positional arguments!'

            for kwarg in kwargs:
                if kwarg in kw_ops:
                    if not kw_ops[kwarg](kwargs[kwarg]):
                        raise PreconditionError('Argument %s failed precondition check' %
                                                (kwarg,))

            with rethrow_as(TypeError, PreconditionError):
                for arg, precond in itertools.zip_longest(args, tn_ops, fillvalue=_TRUE):
                    if precond(arg) is False:
                        raise PreconditionError(
                            'Argument of value %s failed precondition check' % (arg,))
            return fun(*args, **kwargs)

        return inner

    return outer


@for_argument(source_to_function)
def postcondition(condition: Condition):
    """
    Return a decorator, asserting that result of this function, called with provided
    callable,
    is True, else the function will raise PreconditionError.

    :param condition: callable that accepts a single argument, the return value of the function.
        Can be also a string, in which case it is an expression about the value x of return
    """
    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            v = fun(*args, **kwargs)
            if condition(v) is False:
                raise PreconditionError('Condition not true')
            return v

        return inner

    return outer
